/*
    XmlMessageTest is a fully declarative unit testing framework for XML-based message servers.
    It allows developers or testers to create tests by specifying XML messages to send to the server
    and expected results, without the need to develop additional code.
    These tests can be integrated into a build process and run without user interaction.
     
    Copyright (C) 2006 Gene Mitelman
    Email: gene@smartedgellc.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

package com.smartedgellc.xmlmessagetest;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import java.util.regex.*;

public class XmlMessageTest {

	private String testurl;
	private String testsdir;
	private Hashtable globalvalues;
	private int tests, failures;


//--- Main and constructor methods
	
	public static void main(String[] args) {
		if (args.length != 2) {
			logEntry("Usage: java -jar xmlmessagetest.jar <url> <test directory>.");
			System.exit(1);
		} else {
			XmlMessageTest xmltest = new XmlMessageTest(args[0], args[1]);
			if (xmltest.run()) {
				System.exit(0);
			} else {
				System.exit(2);
			}
		}
	}

	XmlMessageTest(String url, String dir) {
		// Initialize
		testurl = url;
		testsdir = dir;
		globalvalues = new Hashtable();
		tests = 0;
		failures = 0;
	}


//--- Main process method

	public boolean run() {
		boolean success = false;
		
		try {
			File testdir = new File(testsdir);
			File[] files = testdir.listFiles(new TestFileFilter());
			for (int i = 0; i < files.length; i++) {
				try {
					processTestFile(files[i]);
				} catch (Exception ex) {
					failures++;
					logEntry("Error running tests: " + ex.toString());
				}
			}
			// Finish up
			if (failures > 0) {
				logEntry("FAILED " + failures + " of " + tests);
			} else {
				logEntry("PASSED " + tests + " of " + tests);
				success = true;
			}
		} catch (Exception ex) {
			logEntry("Error running tests.\n" + ex.toString());
		}
		return success;
	}


//--- Utility methods

	public static void logEntry(String logentry) {
		// Logs any provided information to standard output
		System.out.println(logentry);
	}


//--- Helper methods

	private Hashtable processMessage(String message) throws Exception {
		// Parses out values from passed in message, appends/replaces them in global values table and returns them
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder docbuilder = factory.newDocumentBuilder();
		Document doc = docbuilder.parse(new InputSource(new StringReader(message)));
		Node node = doc.getFirstChild();
		Hashtable results;

		// Set all resulting values
		results = getValues("", node);
		globalvalues.putAll(results);
		
		return results;
	}

	private String replaceValues(String message) throws Exception {
		// Replaces all occurences of <%=value%> in the message with value from the values table
		StringBuffer sb = new StringBuffer();
		String match, valuekey, value;
		
		Pattern pattern = Pattern.compile("<%=.+?%>");
		Matcher matcher = pattern.matcher(message);
		while (matcher.find()) {
			match = matcher.group();
			valuekey = match.substring(3, match.length() - 2);
			if (globalvalues.containsKey(valuekey)) {
				value = (String) globalvalues.get(valuekey);
			} else {
				value = "";
			}
		    matcher.appendReplacement(sb, value);
		}
		matcher.appendTail(sb);
		return sb.toString();
	}

	private Hashtable getValues(String parentname, Node node) throws Exception {
		// Returns a hashtable of all values and attributes in this node and its children
		Hashtable values = new Hashtable();
		String nodename;
		
		// Get value from this node - if any
		if (node.getNodeType() == Node.CDATA_SECTION_NODE || node.getNodeType() == Node.TEXT_NODE) {
			values.put(parentname, node.getNodeValue().trim());
		}
		// Set up node name so can store results for children in the hashtable
		if (parentname.equals("")) {
			nodename = node.getNodeName();
		} else {
			nodename = parentname + "." + node.getNodeName();
		}
		// Get values of all attributes
		NamedNodeMap attributes = node.getAttributes();
		if (attributes != null) {
			for (int i = 0; i < attributes.getLength(); i++) {
				values.put(nodename + "." + attributes.item(i).getNodeName(), attributes.item(i).getNodeValue());
			}
		}
		// Get values of all child nodes
		NodeList nodes = node.getChildNodes();
		for (int i = 0; i < nodes.getLength(); i++) {
			values.putAll(getValues(nodename, nodes.item(i)));
		}

		return values;
	}
	
	private boolean compareValues(String test, Hashtable expectedresults, Hashtable actualresults) {
		// Compares actual and expected results - based on regular expressions
		String expectedvalue, actualvalue, value;
		boolean match;
		
		Enumeration values = expectedresults.keys();
		while (values.hasMoreElements()) {
			value = (String) values.nextElement();
			expectedvalue = (String) expectedresults.get(value);
			match = false;
			actualvalue = (String) actualresults.get(value);
			if (actualvalue != null) {
				if (actualvalue.matches(expectedvalue)) {
					match = true;
				}
			}
			if (!match) {
				logEntry("  Failed '" + test + "': expected " + value + "='" + expectedvalue + "', actual " + value + "='" + actualvalue + "'.");				
				return false;
			}

		}
		
		return true;
	}

	private String sendMessage(String message) throws Exception {
		// Sends message to the server and returns response message
		String line;
		
		message = URLEncoder.encode(message, "UTF-8");
		URL url = new URL(testurl);
		HttpURLConnection httpconn = (HttpURLConnection) url.openConnection();
		httpconn.setRequestMethod("POST");
		httpconn.setRequestProperty("Content-Length", String.valueOf(message.length()));
		httpconn.setDoInput(true);
		httpconn.setDoOutput(true);
		httpconn.connect();
		OutputStream os = httpconn.getOutputStream();
		os.write(message.getBytes());
		os.flush();
		BufferedReader reader = new BufferedReader(new InputStreamReader(httpconn.getInputStream()));
		StringBuffer sb = new StringBuffer();
		while ((line = reader.readLine()) != null) {
			sb.append(line + "\n");
		}
		return sb.toString();
	}
	
	private boolean runTest(String test, String messagein, String expectedmessage) throws Exception {
		// Runs individual test by sending message to the server and comparing actual and expected results
		String messageout, fullmessage;
		Hashtable expectedresults, actualresults = null;
		
		// Send message
		fullmessage = replaceValues(messagein);
		messageout = sendMessage(fullmessage);
		// Identify expected results
		fullmessage = replaceValues(expectedmessage);
		expectedresults = processMessage(fullmessage);
		// Get actual results
		fullmessage = replaceValues(messageout);
		actualresults = processMessage(fullmessage);
		// Compare expected and actual results
		return compareValues(test, expectedresults, actualresults);
	}

	private void processTestFile(File file) throws Exception {
		// Processes each individual test in the specified test file
		String message, messagetype, token, messagein = "", test = "", line, filename = "", filecontents;
		StringTokenizer st;
		StringBuffer sb;
		int index;

		// Read file contents
		filename = file.getName();
		logEntry("Running tests in " + filename);
		BufferedReader br = new BufferedReader(new FileReader(file));
		sb = new StringBuffer();
		while (br.ready()) {
			line = br.readLine();
			sb.append(line + "\n");
		}
		br.close();
		filecontents = sb.toString();
		// Parse out the file and run tests
		st = new StringTokenizer(filecontents, "#");
		while (st.hasMoreTokens()) {
			token = st.nextToken().trim();
			index = token.indexOf("\n");
			if (index >= 0) {
				messagetype = token.substring(0, index).trim();
				message = token.substring(index).trim();
				if (messagetype.startsWith("MESSAGEIN")) {
					messagein = message;
					index = messagetype.indexOf(":");
					if (index >= 0) {
						test = messagetype.substring(index + 1).trim();
					} else {
						test = "";
					}
				} else if (messagetype.startsWith("MESSAGEOUT")) {
					tests++;
					if (!runTest(test, messagein, message)) {
						failures++;
					}
				}
			}
		}
	}
	
}